//Ethan Alexander Shulman 2022 "Flying Through Grass" May 7, 2022

//scene rendering options
#define VIEW_STEPS 60
#define VIEW_STEP_DITHER .1
#define VIEW_EPSILON 1e-4
#define VIEW_EPSILON_MUL 1.12
#define LIGHTING_STEPS 100
#define LIGHTING_STEP_DITHER .1
#define LIGHTING_EPSILON 4e-3
//#define COLOR_MAX 100.
#define DENOISE_RANGE 2
#define DENOISE_NORMAL_THRESHOLD .98
#define DENOISE_DEPTH_THRESHOLD 1e2
#define LOW_DISCREPANCY 1
//#define ANIMATION_TIME 60.
#define ANIMATION_SAMPLES 20
#ifdef ANIMATION_TIME
out float time;
#else
float time = 4.;
#endif

//constants
#define PI 3.14159265



float twave(float x) {
	return 1.-abs(fract(x)-.5)*2.;
}

mat2 r2d(float a) {
	float c = cos(a), s = sin(a);
	return mat2(c,s,-s,c);
}

uvec4 hash(uvec4 seed) {
    uvec4 h = (0x6A7F8FAAu^seed)*0x01000193u;
    h = ((h.wxyz>>3u)^h^seed.yzwx)*0x01000193u;
    h = ((h.zwxy>>8u)^h^seed.wxyz)*0x01000193u;
    return h^(h>>11u);
}
#define I2F (1./float(0xFFFFFFFFu))

float vvoronoi(vec3 p, vec3 s, float m) {
	return min(length(fract(1000.+p+cos(s)*m)-.5),length(fract(1000.+p-.5+sin(s)*m)-.5))*1.15;
}

#define grassPath(p) cos(p.zx*vec2(2.19823,3.87274)-p.y*2.18723)*.1

/*returns camera data as mat3 in the following structure {
vec3 position;
float roll, pitch, yaw;
float fieldOfViewTangent, viewNear, orbitDistance;*/
mat3 camera() {
	float et = time, height = twave(et/60.);
	vec3 cp = vec3(0,.03+pow(height,2.),et*.1);
	cp.xz -= grassPath(cp);
	return mat3(cp,
			0,-height,0,
			1.,0,0);
}


//distance dithering for volumetrics
#define START_VOLUME(posSeed) vec2 opacityRand = fract(baseRand.zw+128.*dot(abs(posSeed),baseRand.zxy)*baseRand.xy)
#define VOLUME(dst,skip,alpha,anisotropy) max(dst,-1e3+(skip+1e3)*step(alpha+nsampling*anisotropy,opacityRand.x))

//dither materials for chromatic separation
#define START_CHROMA(posSeed) float chromaFrame = mod(baseRand.x+156.*dot(abs(posSeed),baseRand.zwy),3.)
#define CHROMA(rc,gc,bc) if (chromaFrame < 1.) {rc;}\
else if (chromaFrame < 2.) {gc;}\
else {bc;}



//functions defining scene
//returns vec2(distance, materialId) of nearest surface
vec2 geometry(vec3 p) {
	if (p.y > 1.3) return vec2(p.y-.9,1);

	//dirt
	vec2 geo = vec2(p.y,0);
	if (geo.x < .4) {
		geo.x += -vvoronoi(p*10.,p.yzx*.062434*vec3(27,36,42),1.)*.05+
				vvoronoi(p*33.,p.zxy*.1873*vec3(111,78,90),1.)*.02+
				vvoronoi(p*100.,p.yzx*.07654*vec3(103,111,100),1.)*.01;
	}
	
	//grass
	vec3 gp;
	gp.xz = p.xz+grassPath(p);//cos(p.zx*vec2(2.19823,3.87274)-p.y*2.18723)*.1;
	vec2 fp = floor(abs(gp.xz)/.06);
	gp.y = p.y;
	gp.xz = mod(abs(gp.xz),.06)-.03;
	gp.xz *= r2d(fp.x*.23764+fp.y*.9829634);
	float gd = length(max(abs(gp)-vec3(.01,.5,.001),0.));
	if (gd < geo.x) geo = vec2(gd,1);

	return geo;
}

/*returns material as mat3 in the following structure {
vec3 color;
float roughness, emission, opacity;
float refractionIndex;*/
mat3 material(vec3 p, vec3 n, vec3 rd, float mat) {
	if (mat == 1.) {
		//grass
		if (baseRand.w < .1) return mat3(vec3(.8), .00004,0,1, 1,0,0);//gloss
		return mat3(vec3(.004,.634,.14), .4,0,1, 1,0,0);
	}

	//dirt
	return mat3(vec3(1.,.35,.1)*.3, .3,0,1, 1,0,0);
}

//returns background color at point and direction
vec3 background(vec3 p, vec3 d) {
	return mix(vec3(1),vec3(.3,.65,1),pow(max(0.,d.y),.5));
}
vec3 backgroundLighting(vec3 p, vec3 d) {
	return background(p,d)*max(0.,d.y)*.2+vec3(1,.98,.95)*100.*max(0.,1.-4.*length(d-normalize(vec3(1,3,2))));
}

